from typing import (
    Any,
    Dict,
    Generic,
    Iterable,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from pydantic import BaseModel
from pymongo import UpdateOne
from pymongo.collection import Collection
from pymongo.database import Database
from pymongo.results import InsertOneResult, UpdateResult

from .pagination import (
    Edge,
    decode_pagination_cursor,
    encode_pagination_cursor,
    get_pagination_cursor_payload,
)

T = TypeVar("T", bound=BaseModel)
OutputT = TypeVar("OutputT", bound=BaseModel)

Sort = Sequence[Tuple[str, int]]


class AbstractRepository(Generic[T]):
    class Meta:
        collection_name: str

    def __init__(self, database: Database):
        super().__init__()
        self.__database: Database = database
        self.__document_class = (
            getattr(self.Meta, "document_class")
            if hasattr(self.Meta, "document_class")
            else self.__orig_bases__[0].__args__[0]  # type: ignore
        )
        self.__collection_name = self.Meta.collection_name
        self.__validate()

    """
    Get pymongo collection
    """

    def get_collection(self) -> Collection:
        return self.__database[self.__collection_name]

    def __validate(self):
        if not issubclass(self.__document_class, BaseModel):
            raise Exception("Document class should inherit BaseModel")
        if "id" not in self.__document_class.model_fields:
            raise Exception("Document class should have id field")
        if not self.__collection_name:
            raise Exception("Meta should contain collection name")

    @staticmethod
    def to_document(model: T) -> dict:
        """
        Convert model to document
        :param model:
        :return: dict
        """
        data = model.model_dump()
        data.pop("id")
        if model.id:
            data["_id"] = model.id
        return data

    def __map_id(self, data: dict) -> dict:
        query = data.copy()
        if "id" in data:
            query["_id"] = query.pop("id")
        return query

    def __map_sort(self, sort: Sort) -> Optional[Sort]:
        result = []
        for item in sort:
            key = item[0]
            ordering = item[1]
            if key == "id":
                key = "_id"
            result.append((key, ordering))
        return result

    def to_model_custom(self, output_type: Type[OutputT], data: dict) -> OutputT:
        """
        Convert document to model with custom output type
        """
        data_copy = data.copy()
        if "_id" in data_copy:
            data_copy["id"] = data_copy.pop("_id")
        return output_type.model_validate(data_copy)

    def to_model(self, data: dict) -> T:
        """
        Convert document to model
        """
        return self.to_model_custom(self.__document_class, data)

    def save(self, model: T) -> Union[InsertOneResult, UpdateResult]:
        """
        Save entity to database. It will update the entity if it has id, otherwise it will insert it.
        """
        document = self.to_document(model)

        if model.id:
            mongo_id = document.pop("_id")
            return self.get_collection().update_one(
                {"_id": mongo_id}, {"$set": document}, upsert=True
            )

        result = self.get_collection().insert_one(document)
        model.id = result.inserted_id
        return result

    def save_many(self, models: Iterable[T]):
        """
        Save multiple entities to database
        """
        models_to_insert = []
        models_to_update = []

        for model in models:
            if model.id:
                models_to_update.append(model)
            else:
                models_to_insert.append(model)
        if len(models_to_insert) > 0:
            result = self.get_collection().insert_many(
                (self.to_document(model) for model in models_to_insert)
            )

            for idx, inserted_id in enumerate(result.inserted_ids):
                models_to_insert[idx].id = inserted_id

        if len(models_to_update) == 0:
            return

        documents_to_update = [self.to_document(model) for model in models_to_update]
        mongo_ids = [doc.pop("_id") for doc in documents_to_update]
        bulk_operations = [
            UpdateOne({"_id": mongo_id}, {"$set": document}, upsert=True)
            for mongo_id, document in zip(mongo_ids, documents_to_update)
        ]
        self.get_collection().bulk_write(bulk_operations)

    def delete(self, model: T):
        return self.get_collection().delete_one({"_id": model.id})

    def find_one_by_id(self, _id: Any) -> Optional[T]:
        """
        Find entity by id

        Note: The id should be of the same type as the id field in the document class, ie. ObjectId
        """
        return self.find_one_by({"id": _id})

    def find_one_by(self, query: dict) -> Optional[T]:
        """
        Find entity by mongo query
        """
        result = self.get_collection().find_one(self.__map_id(query))
        return self.to_model(result) if result else None

    def find_by_with_output_type(
        self,
        output_type: Type[OutputT],
        query: dict,
        skip: Optional[int] = None,
        limit: Optional[int] = None,
        sort: Optional[Sort] = None,
        projection: Optional[Dict[str, int]] = None,
    ) -> Iterable[OutputT]:
        """
        Find entities by mongo query allowing custom output type
        :param output_type:
        :param query:
        :param skip:
        :param limit:
        :param sort:
        :param projection:
        :return:
        """
        mapped_projection = self.__map_id(projection) if projection else None
        mapped_sort = self.__map_sort(sort) if sort else None
        cursor = self.get_collection().find(self.__map_id(query), mapped_projection)
        if limit:
            cursor.limit(limit)
        if skip:
            cursor.skip(skip)
        if sort:
            cursor.sort(mapped_sort)
        return map(lambda doc: self.to_model_custom(output_type, doc), cursor)

    def find_by(
        self,
        query: dict,
        skip: Optional[int] = None,
        limit: Optional[int] = None,
        sort: Optional[Sort] = None,
        projection: Optional[Dict[str, int]] = None,
    ) -> Iterable[T]:
        """ "
        Find entities by mongo query
        """
        return self.find_by_with_output_type(
            output_type=self.__document_class,
            query=query,
            skip=skip,
            limit=limit,
            sort=sort,
            projection=projection,
        )

    def get_pagination_query(
        self,
        query: dict,
        after: Optional[str] = None,
        before: Optional[str] = None,
        sort: Optional[Sort] = None,
    ) -> dict:
        """
        Build pagination query based on the cursor and sort
        """
        generated_query: dict = {"$and": [query]}
        selected_cursor = after or before

        if selected_cursor and sort:
            cursor_data = decode_pagination_cursor(selected_cursor)
            dict_values = []
            for i, sort_expression in enumerate(sort):
                if after:
                    compare_operator = "$gt" if sort_expression[1] > 0 else "$lt"
                else:
                    compare_operator = "$lt" if sort_expression[1] > 0 else "$gt"
                dict_values.append(
                    (sort_expression[0], {compare_operator: cursor_data[i]})
                )
            generated_query["$and"].append(dict(dict_values))

        if len(generated_query["$and"]) == 1:
            generated_query = query or {}

        return generated_query

    def paginate_with_output_type(
        self,
        output_type: Type[OutputT],
        query: dict,
        limit: int,
        after: Optional[str] = None,
        before: Optional[str] = None,
        sort: Optional[Sort] = None,
        projection: Optional[Dict[str, int]] = None,
    ) -> Iterable[Edge[OutputT]]:
        """
        Paginate entities by mongo query allowing custom output type
        """
        sort_keys = []

        if not sort:
            sort = [("_id", 1)]

        for sort_expression in sort:
            sort_keys.append(sort_expression[0])

        models = self.find_by_with_output_type(
            output_type,
            query=self.get_pagination_query(
                query=query, after=after, before=before, sort=sort
            ),
            limit=limit,
            sort=sort,
            projection=projection,
        )

        return map(
            lambda model: Edge[T](
                node=model,
                cursor=encode_pagination_cursor(
                    get_pagination_cursor_payload(model, sort_keys)
                ),
            ),
            models,
        )

    def paginate(
        self,
        query: dict,
        limit: int,
        after: Optional[str] = None,
        before: Optional[str] = None,
        sort: Optional[Sort] = None,
        projection: Optional[Dict[str, int]] = None,
    ) -> Iterable[Edge[T]]:
        """
        Paginate entities by mongo query using cursor based pagination

        Return type is an iterable of Edge objects, which contain the model and the cursor
        """
        return self.paginate_with_output_type(
            self.__document_class,
            query,
            limit,
            after=after,
            before=before,
            sort=sort,
            projection=projection,
        )



from typing import Any

from bson import ObjectId
from pydantic_core import core_schema


class ObjectIdField(ObjectId):
    @classmethod
    def __get_pydantic_core_schema__(
        cls, _source_type: Any, _handler: Any
    ) -> core_schema.CoreSchema:
        object_id_schema = core_schema.chain_schema(
            [
                core_schema.str_schema(),
                core_schema.no_info_plain_validator_function(cls.validate),
            ]
        )
        return core_schema.json_or_python_schema(
            json_schema=object_id_schema,
            python_schema=core_schema.union_schema(
                [core_schema.is_instance_schema(ObjectId), object_id_schema]
            ),
            serialization=core_schema.plain_serializer_function_ser_schema(
                lambda x: str(x)
            ),
        )

    @classmethod
    def validate(cls, value):
        if not ObjectId.is_valid(value):
            raise ValueError("Invalid id")

        return ObjectId(value)



from typing import List

import mongomock
import pytest
from bson import ObjectId
from pydantic import BaseModel, Field

from pydantic_mongo import AbstractRepository, ObjectIdField
from pydantic_mongo.errors import PaginationError


class Foo(BaseModel):
    count: int
    size: float = None


class Bar(BaseModel):
    apple: str = Field(default="x")
    banana: str = Field(default="y")


class Spam(BaseModel):
    id: ObjectIdField = None
    foo: Foo = None
    bars: List[Bar] = None


class SpamRepository(AbstractRepository[Spam]):
    class Meta:
        collection_name = "spams"


@pytest.fixture
def database():
    return mongomock.MongoClient().db


class TestRepository:
    def test_save(self, database):
        spam_repository = SpamRepository(database=database)
        foo = Foo(count=1, size=1.0)
        bar = Bar()
        spam = Spam(foo=foo, bars=[bar])
        spam_repository.save(spam)

        assert {
            "_id": ObjectId(spam.id),
            "foo": {"count": 1, "size": 1.0},
            "bars": [{"apple": "x", "banana": "y"}],
        } == database["spams"].find()[0]

        spam.foo.count = 2
        spam_repository.save(spam)

        assert {
            "_id": ObjectId(spam.id),
            "foo": {"count": 2, "size": 1.0},
            "bars": [{"apple": "x", "banana": "y"}],
        } == database["spams"].find()[0]

    def test_save_upsert(self, database):
        spam_repository = SpamRepository(database=database)
        spam = Spam(
            id=ObjectId("65012da68ea5a4798502f710"),
            foo=Foo(count=1, size=1.0),
            bars=[]
        )
        spam_repository.save(spam)

        assert {
            "_id": ObjectId(spam.id),
            "foo": {"count": 1, "size": 1.0},
            "bars": [],
        } == database["spams"].find()[0]

    def test_save_many(self, database):
        spam_repository = SpamRepository(database=database)
        spams = [
            Spam(),
            Spam(id=ObjectId("65012da68ea5a4798502f710")),
        ]
        spam_repository.save_many(spams)

        initial_data = [
            {
                "_id": ObjectId(spams[0].id),
                "foo": None,
                "bars": None,
            },
            {
                "_id": ObjectId(spams[1].id),
                "foo": None,
                "bars": None,
            },
        ]

        assert initial_data == list(database["spams"].find())

        # Calling save_many again will only update
        spam_repository.save_many(spams)
        assert initial_data == list(database["spams"].find())

        # Calling save_many with only a new model will only insert
        new_span = Spam()
        spam_repository.save_many([new_span])
        assert new_span.id is not None
        assert 3 == database["spams"].count_documents({})

    def test_delete(self, database):
        spam_repository = SpamRepository(database=database)
        foo = Foo(count=1, size=1.0)
        bar = Bar()
        spam = Spam(foo=foo, bars=[bar])
        spam_repository.save(spam)

        result = spam_repository.find_one_by_id(spam.id)
        assert result is not None

        spam_repository.delete(spam)
        result = spam_repository.find_one_by_id(spam.id)
        assert result is None

    def test_find_by_id(self, database):
        spam_id = ObjectId("611827f2878b88b49ebb69fc")
        database.spams.insert_one(
            {
                "_id": spam_id,
                "foo": {"count": 2, "size": 1.0},
                "bars": [{"apple": "x", "banana": "y"}],
            }
        )

        spam_repository = SpamRepository(database=database)
        result = spam_repository.find_one_by_id(spam_id)

        assert issubclass(Spam, type(result))
        assert spam_id == result.id
        assert "x" == result.bars[0].apple

    def test_find_by(self, database):
        database.spams.insert_many(
            [
                {
                    "foo": {"count": 2, "size": 1.0},
                    "bars": [{"apple": "x", "banana": "y"}],
                },
                {
                    "foo": {"count": 3, "size": 1.0},
                    "bars": [{"apple": "x", "banana": "y"}],
                },
            ]
        )

        spam_repository = SpamRepository(database=database)

        # Simple Find
        result = spam_repository.find_by({})
        results = [x for x in result]
        assert 2 == len(results)
        assert 2 == results[0].foo.count
        assert 3 == results[1].foo.count

        # Find with optional parameters
        result = spam_repository.find_by(
            {}, skip=10, limit=10, sort=[("foo.count", 1), ("id", 1)]
        )
        results = [x for x in result]
        assert 0 == len(results)

    def test_invalid_model_class(self, database):
        class BrokenRepository(AbstractRepository[int]):
            class Meta:
                collection_name = "spams"

        with pytest.raises(Exception):
            BrokenRepository(database=database)

    def test_invalid_model_id_field(self, database):
        class NoIdModel(BaseModel):
            something: str

        class BrokenRepository(AbstractRepository[NoIdModel]):
            class Meta:
                collection_name = "spams"

        with pytest.raises(Exception):
            BrokenRepository(database=database)

    def test_invalid_model_collection_name(self, database):
        class BrokenRepository(AbstractRepository[Spam]):
            class Meta:
                collection_name = None

        with pytest.raises(Exception):
            BrokenRepository(database=database)

    def test_paginate(self, database):
        database.spams.insert_many(
            [
                {
                    "_id": ObjectId("611b140f4eb6ee47e966860f"),
                    "foo": {"count": 2, "size": 1.0},
                    "bars": [{"apple": "x", "banana": "y"}],
                },
                {
                    "id": ObjectId("611b141cf533ca420b7580d6"),
                    "foo": {"count": 3, "size": 1.0},
                    "bars": [{"apple": "x", "banana": "y"}],
                },
                {
                    "_id": ObjectId("611b15241dea2ee3f7cbfe30"),
                    "foo": {"count": 2, "size": 1.0},
                    "bars": [{"apple": "x", "banana": "y"}],
                },
                {
                    "_id": ObjectId("611b157c859bde7de88c98ac"),
                    "foo": {"count": 2, "size": 1.0},
                    "bars": [{"apple": "x", "banana": "y"}],
                },
                {
                    "_id": ObjectId("611b158adec89d18984b7d90"),
                    "foo": {"count": 2, "size": 1.0},
                    "bars": [{"apple": "x", "banana": "y"}],
                },
            ]
        )

        spam_repository = SpamRepository(database=database)

        # Simple Find
        result = list(spam_repository.paginate({}, limit=10))
        assert len(result) == 5

        # Find After
        result = list(
            spam_repository.paginate(
                {}, limit=10, after="eNqTYWBgYCljEAFS7AYMidKiXfdOzJWY4V07gYEBAD7HBkg="
            )
        )
        assert len(result) == 1

        # Find Before
        result = list(
            spam_repository.paginate(
                {}, limit=10, before="eNqTYWBgYCljEAFS7AYMidKiXfdOzJWY4V07gYEBAD7HBkg="
            )
        )
        assert len(result) == 3

        with pytest.raises(PaginationError):
            spam_repository.paginate({}, limit=10, after="invalid string")



import datetime
from typing import List

from bson import ObjectId
from pydantic import BaseModel

from pydantic_mongo.pagination import (
    decode_pagination_cursor,
    encode_pagination_cursor,
    get_pagination_cursor_payload,
)


class Foo(BaseModel):
    count: int
    size: float = None


class Bar(BaseModel):
    apple: str = "x"
    banana: str = "y"


class Spam(BaseModel):
    id: str = None
    foo: Foo
    bars: List[Bar]


class TestPagination:
    def test_get_pagination_cursor_payload(self):
        spam = Spam(id="lala", foo=Foo(count=1, size=1.0), bars=[Bar()])

        values = get_pagination_cursor_payload(spam, ["_id", "id"])
        assert values[0] == "lala"
        assert values[1] == "lala"

        values = get_pagination_cursor_payload(spam, ["foo.count"])
        assert values[0] == 1

        values = get_pagination_cursor_payload(spam, ["bars.0.apple"])
        assert values[0] == "x"

    def test_cursor_encoding(self):
        old_value = [ObjectId("611b158adec89d18984b7d90"), "a", 1]
        cursor = encode_pagination_cursor(old_value)
        new_value = decode_pagination_cursor(cursor)
        assert old_value == new_value



import pytest
from bson import ObjectId
from pydantic import BaseModel, ValidationError

from pydantic_mongo import ObjectIdField


class User(BaseModel):
    id: ObjectIdField = None


class TestFields:
    def test_object_id_validation(self):
        with pytest.raises(ValidationError):
            User.model_validate({"id": "lala"})
        User.model_validate({"id": "611827f2878b88b49ebb69fc"})

    def test_object_id_serialize(self):
        lala = User(id=ObjectId("611827f2878b88b49ebb69fc"))
        json_result = lala.model_dump_json()
        assert '{"id":"611827f2878b88b49ebb69fc"}' == json_result

    def test_modify_schema(self):
        user = User(id=ObjectId("611827f2878b88b49ebb69fc"))
        schema = user.model_json_schema()
        assert {
            "title": "User",
            "type": "object",
            "properties": {"id": {"default": None, "title": "Id", "type": "string"}},
        } == schema



import pytest
from pydantic import BaseModel, Field

from pydantic_mongo import AbstractRepository, ObjectIdField


class HamModel(BaseModel):
    id: ObjectIdField = Field(default=None)
    name: str


class HamRepository(AbstractRepository):
    class Meta:
        collection_name = "ham"
        document_class = HamModel


@pytest.fixture(scope="session")
def ham_repo(database):
    return HamRepository(database=database)


@pytest.fixture
def clean_ham_collection(ham_repo):
    return ham_repo.get_collection().delete_many({})


def test_repository_with_v2_meta(ham_repo):
    assert not list(ham_repo.find_by({})), "should have no documents in db"
    assert ham_repo.get_collection().name == "ham"


def test_save_with_new_repo(clean_ham_collection, ham_repo):
    m = HamModel(name="wilfred")
    assert m.id is None, "should have no id"
    ham_repo.save(m)
    assert m.id



import os

from setuptools import setup

long_description = open("README.md", "r").read()

package_root = os.path.abspath(os.path.dirname(__file__))

version = {}
with open(os.path.join(package_root, "pydantic_mongo/version.py")) as fp:
    exec(fp.read(), version)
version = version["__version__"]

setup(
    name="pydantic-mongo",
    version=version,
    packages=["pydantic_mongo"],
    setup_requires=["wheel"],
    install_requires=["pymongo>=4.3,<5.0", "pydantic>=2.0.2,<3.0.0"],
    description="Document object mapper for pydantic and pymongo",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jefersondaniel/pydantic-mongo",
    author="Jeferson Daniel",
    author_email="jeferson.daniel412@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    python_requires=">=3.8",
)



import zlib
from base64 import b64decode, b64encode
from typing import Any, Generic, List, TypeVar

import bson
from pydantic import BaseModel

from .errors import PaginationError

DataT = TypeVar("DataT")


class Edge(BaseModel, Generic[DataT]):
    node: DataT
    cursor: str


def encode_pagination_cursor(data: List) -> str:
    byte_data = bson.BSON.encode({"v": data})
    byte_data = zlib.compress(byte_data, 9)
    return b64encode(byte_data).decode("utf-8")


def decode_pagination_cursor(data: str) -> List:
    try:
        byte_data = b64decode(data.encode("utf-8"))
        byte_data = zlib.decompress(byte_data)
        result = bson.BSON(byte_data).decode()
        return result["v"]
    except Exception:
        raise PaginationError("Invalid cursor")


def get_pagination_cursor_payload(model: BaseModel, keys: List[str]) -> List[Any]:
    model_dict = model.model_dump()
    model_dict["_id"] = model_dict["id"]

    return [__evaluate_dot_notation(model_dict, key) for key in keys]


def __evaluate_dot_notation(data: Any, path: str):
    pieces = path.split(".")

    if len(pieces) == 1:
        return data[path]

    current_data = data

    for piece in pieces:
        if isinstance(current_data, list) or isinstance(current_data, tuple):
            current_data = current_data[int(piece)]
        else:
            current_data = current_data[piece]

    return current_data
